/* Core library for the Barons strategic engine */

#include "core.h"

/* Define extern consts from baron.h */
const char thing_name[tt_count][16] = {"Unknown", "ResourceT", "Resource",
                     "Race", "Population", "FactoryT", "Factory", "SiteT",
                     "Site", "Allocation", "Construction", "Territory"};
const char thing_abbr[tt_count] = {'?', 'R', 'r', 'P', 'p', 'F', 'f',
                                   'S', 's', 'a', 'c', 'T'};

static thing_data *res_skill, *res_happiness;

/* Root pointers for the various lists of Things, locked away where other
   modules won't be able to see them, use them, or cause problems when the
   Thing storage structures change */
static thing_data *master_list[tt_count] = {0};

/* ---------------------------------------------------------------------- */

/* Thing list iterators */
thing_data *next_thing(const thing_data *thing) {
  return (thing->next);
}

thing_data *prev_thing(const thing_data *thing) {
  return (thing->prev);
}

/* ---------------------------------------------------------------------- */

/* General Thing list manipulation functions */

int add_to_list(thing_data **list_root, thing_data *thing) {
  /* In an attempt to reduce later search times, even if only slightly, we're
     inserting Things into lists in alphabetical order.  Note that the
     insertion will fail if a Thing is already present in the list with the
     same name! */
  thing_data *curr, *last;
  char thing_name[16];
  int namecmp;

  /* Things not allowed to have empty names */
  if (!thing->data.generic.name[0]) return 0;

  /* Handle empty lists right off */
  if (!*list_root) {
    *list_root = thing;
    thing->next = thing->prev = 0;

    return 1;
  }

  /* Iterate over the list until we find the correct position for *thing */
  curr = *list_root;
  strcpy(thing_name, thing->data.generic.name);
  while ((namecmp = strcmp(thing_name, curr->data.generic.name)) > 0) {
    if (curr->next) {
      last = curr;
      curr = curr->next;
    } else {    /* At end of list, so add *thing to the tail */
      curr->next = thing;
      thing->prev = curr;
      thing->next = 0;

      return 1;
    }
  }

  /* If the Thing is already in this list, fail */
  if (!namecmp) return 0;

  if (curr == *list_root) {     /* Inserting at head */
    thing->prev = 0;
    thing->next = curr;
    curr->prev = thing;
    *list_root = thing;
  } else {                      /* Inserting mid-list */
    thing->prev = last;
    thing->next = curr;
    curr->prev = thing;
    last->next = thing;
  }

  return 1;
}

int remove_from_list(thing_data **list_root, thing_data *thing) {
  thing_data **pred_ptr;

  /* If there's a previous node, update its next ptr, otherwise we're at the
     head of the list and will need to update *list_root */
  pred_ptr = (thing->prev ? &(thing->prev->next) : list_root);
  *pred_ptr = thing->next;

  if (thing->next) thing->next->prev = thing->prev;

  return 1;
}

thing_data *find_in_list(const thing_data *list_root, const char* name) {
  const thing_data *curr = list_root;
  int namecmp;

  /* Empty string matches first Thing in the list */
  if (!name[0]) return (thing_data*)list_root;

  while (curr) {
    namecmp = strcmp(name, curr->data.generic.name);
    if (!namecmp)             /* Found it! */
      return (thing_data*)curr;
    else if (namecmp < 0)     /* List is sorted, so we know it's not here */
      return 0;

    curr = next_thing(curr);
  }

  return 0;
}

/* ---------------------------------------------------------------------- */

/* Master Thing list manipulation functions */

int add_to_master(thing_data **thing) {
  thing_data *thing_shell;
  
  /* If there's not a tt_unknown with *thing's name waiting to be defined,
     add the Thing to the appropriate master list */
  thing_shell = find_in_list(master_list[tt_unknown],
                             (*thing)->data.generic.name);
  if (!thing_shell)
    return add_to_list(&master_list[(*thing)->type], *thing);

  /* OK, this Thing has already been referenced, so we'll need to copy its
     data into the existing tt_unknown Thing and free *thing's memory. */
  remove_from_list(&master_list[tt_unknown], thing_shell);
  memcpy(thing_shell, *thing, sizeof(thing_data));
  free(*thing);
  *thing = thing_shell;

  return add_to_list(&master_list[(*thing)->type], *thing);
}

int remove_from_master(thing_data *thing) {
  return remove_from_list(&master_list[thing->type], thing);
}

thing_data *find_thing(const thing_type type, const char* name) {
  return find_in_list(master_list[type], name);
}

thing_data *find_any_thing(const char* name) {
  /* Since we don't know what kind of Thing we're looking for, we'll have to
     go through and check all the master lists.  The order is arbitrary, but
     _should_ only affect performance, because there aren't supposed to be
     Things with the same name, even if they're of different types (except
     for instances with default names, but they shouldn't be in the master
     lists anyhow). */
  thing_data *found_thing;
  thing_data *new_thing;
  thing_type type;

  /* Catch pseudo-things Skill and Happiness */
  if (!strcmp(name, "Skill")) {
    if (!res_skill) {
      res_skill = (thing_data*)calloc(1, sizeof(thing_data));

      res_skill->type = tt_resource_template;
      strcpy(res_skill->data.resource_template.name, "Skill");
    }
    return res_skill;
  }

  if (!strcmp(name, "Happiness")) {
    if (!res_happiness) {
      res_happiness = (thing_data*)calloc(1, sizeof(thing_data));

      res_happiness->type = tt_resource_template;
      strcpy(res_happiness->data.resource_template.name, "Happiness");
    }
    return res_happiness;
  }

  /* Iterate over the lists of Things, checking each type in turn */
  for (type = tt_min ; type <= tt_max ; type++) {
    found_thing = find_in_list(master_list[type], name);
    if (found_thing) return found_thing;
  }

  /* Didn't find it, so create a new Thing of type tt_unknown and return it */
  new_thing = (thing_data*)calloc(1, sizeof(thing_data));
  if (!new_thing) return 0;     /* Allocation failed! */

  new_thing->type = tt_unknown;
  strcpy(new_thing->data.generic.name, name);
  add_to_list(&master_list[tt_unknown], new_thing);

  return new_thing;
}
