/* Barons strategic combat/economic/production engine core definitions    */

#ifndef __BARON_CORE_H__
#define __BARON_CORE_H__

/* Memory management:
   Things are generally allocated dynamically and should always be handled
   using calloc (or malloc) and free, as find_any_thing and add_to_master
   will occasionally create and destroy Things to allow forward references in
   data files.  All Things created by find_any_thing will be of type
   tt_unknown, while add_to_master will only free Things when they are
   initially added to the master lists (after copying their contents to an
   existing tt_unknown Thing). */

/* Forward declaration */
struct thing_struct;

/* Generic */
/* All Thing types start with a char[32] name.  This struct has only a name
   and is intended for use when we don't know or don't care what type of
   Thing is being examined, we just want its name. */
typedef struct {
  char name[32];
} generic_thing_data;

/* Resource */
/* Represents the supply of a given substance in an area.  From the user's/
   conceptual point of view, all resource quantities are scaled by a factor
   of res_scale to become fixed-point numbers. */
enum {res_scale = 1024};
typedef enum { dt_late = 0, dt_early } decay_type;

typedef struct {
  char name[32];

  struct thing_struct *influence;       /* Effects of resource's presence */

  long can_overspend;                      /* Factor for deficit spending */
 
  char decay_rate;          /* Percent that decays each turn - NOT SCALED */
  decay_type decay_when;                    /* When in turn decay happens */
  struct thing_struct *decay_to;     /* What decay produces per unit lost */
} resource_template_data;

typedef struct {
  char name[32];         /* Same as resource->name - here for consistency */

  resource_template_data *type;

  long qty;
  long old_qty;
} resource_data;

/* Race */
/* Base characteristics of a race of creatures.  For all practical purposes,
   this is a Population template. */
typedef struct {
  char name[32];
  
  long food_consumption;                /* Just about everyone needs food */
  long mana_consumption;                         /* For magical creatures */
  struct thing_struct *resource_consumption;     /* For strange appetites */

  long might_production;                                /* Combat ability */
  long labor_production;                      /* Ability to run Factories */
  long knowledge_production;              /* Magical and research ability */
  struct thing_struct *resource_production;              /* Anything else */

  short growth_rate;            /* Could be population_production, but... */
  short wanderlust;             /* Controls migration and hero appearance */
} race_data;

/* Population */
/* Information about the creatures of a specific race living in an area */
typedef struct {
  char name[32];      /* Usually same as race, but can be used for unique */
                                                /* clan/tribe names, etc. */
  race_data *race;

  long size;                               /* Number of creatures present */
  long skill;                                    /* How talented they are */
  long happiness;                                   /* How happy they are */
} population_data;

/* Factory */
/* Generic term for any facility which transforms one or more input Resources
   into one or more output Resources */
typedef struct {
  char name[32];

  struct thing_struct *input;                   /* Raw materials consumed */
  struct thing_struct *output;                 /* Finished goods produced */
  char capacity;                  /* Number of input->output transactions */
                                        /* possible per turn - NOT SCALED */

  struct thing_struct *required;    /* What has to be there to build this */
  struct thing_struct *cost;       /* What you have to spend to build one */
  struct thing_struct *repair_cost;   /* What you have to spend to fix it */

  long depletion;                         /* Chance of capacity reduction */
} factory_template_data;

typedef struct {
  char name[32];

  factory_template_data *type;

  char qty;       /* How many factories of this type are here - NOT SCALED*/
  short capacity;
} factory_data;

/* Site */
/* A special feature of an area, generally one which will have an effect
   on Resources or Factories there */
typedef struct {
  char name[32];

  struct thing_struct *affects;           /* The Things affected by Sites */
  long magnitude;                    /* Multiplier to each affected Thing */
} site_template_data;

typedef struct {
  char name[32];

  site_template_data *type;
} site_data;

/* Allocation */
/* Data on allocation of resources within a Territory */
enum allocation_mode_enum { al_quantity, al_percent, al_max };
typedef enum allocation_mode_enum allocation_mode;
/* al_quantity - Allocate a fixed amount each turn
   al_percent  - Allocate a percentage of what's on hand
   al_max      - Allocate as much as is available, including maximum
                 overspending
*/

typedef struct {
  char name[32];         /* Name of the Thing supplied by this allocation */

  struct thing_struct *resource;            /* Resource that is allocated */
  allocation_mode type;     /* Is amount a percentage, raw quantity, etc? */
} allocation_data;

/* Construction */
/* Information on any construction in progress within a territory */
typedef struct {
  char name[32];

  factory_template_data *factory;    /* Factory under construction */
  struct thing_struct *remaining_cost;  /* List of resources still needed */
} construction_data;

/* Territory */
/* An area where Resources, Populations, Factories, and Sites can be found */
typedef struct {
  char name[32];

  /* Lists of what all is present in this territory */
  struct thing_struct *populace;
  struct thing_struct *sites;
  struct thing_struct *factories;
  struct thing_struct *resources;
  struct thing_struct *allocated;
  struct thing_struct *building;
} territory_data;

/* Thing */
/* Union-containing struct that can represent any of the above, to allow
   for a vaguely object-oriented approach to Things.  Also provides for
   construction of generic linked lists. */
enum thing_type_enum {
  tt_unknown, tt_resource_template, tt_resource, tt_race, tt_population,
  tt_factory_template, tt_factory, tt_site_template, tt_site, tt_allocation,
  tt_construction, tt_territory
};
typedef enum thing_type_enum thing_type;
enum {tt_min = tt_unknown, tt_max = tt_territory, tt_count = tt_max + 1};
extern const char thing_name[tt_count][16];
extern const char thing_abbr[tt_count];

typedef struct thing_struct {
  thing_type type;
  union {
    generic_thing_data generic;
    resource_template_data resource_template;
    resource_data resource;
    race_data race;
    population_data population;
    factory_template_data factory_template;
    factory_data factory;
    site_template_data site_template;
    site_data site;
    allocation_data allocation;
    construction_data construction;
    territory_data territory;
  } data;

  /* DO NOT rely on the presence of *next_thing or the assumption that Things
     will always live in doubly-linked lists!  Once we get more than a couple
     hundred Things in the world, the data structure they live in will
     probably need to change, so just assume that it could change at any
     time. */
  struct thing_struct *next, *prev;
} thing_data;

/* Skill and Happiness function differently enough from other Resources that
   they've been hardcoded. */
extern thing_data *res_skill, *res_happiness;

/* ---------------------------------------------------------------------- */

/* Things generally live in linked lists, so we'll need some basic list
   maintenance functions.  We could make a single list for Things of all
   types, but that would substantially increase the amount of work that
   would have to be done for searches. */

/* Functions to deal with the master lists of Things (will generally hold
   templates and Territories only) */
extern int add_to_master(thing_data **thing);
extern int remove_from_master(thing_data *thing); /* Does not free memory! */
extern thing_data *find_thing(const thing_type type, const char* name);
extern thing_data *find_any_thing(const char* name);

/* Functions to deal with lists of Things belonging to another Thing.  These
   lists are generally of a single type, so find_in_list doesn't have any use
   for a type parameter, unlike find_thing above.  These lists will generally
   not hold templates (but may be lists of Territories). */
extern int add_to_list(thing_data **list_root, thing_data *thing);
extern int remove_from_list(thing_data **list_root, thing_data *thing);
extern thing_data *find_in_list(const thing_data *list_root, const char* name);

/* And a couple functions for walking through lists of Things */
extern thing_data *next_thing(const thing_data *thing);
extern thing_data *prev_thing(const thing_data *thing);

#endif
