/* Basic console display routines for Barons.  Geared towards spitting out
   data in a human-friendly format rather than the machine-parsable version
   generated by the functions in file.c */

#include <stdio.h>
#include "core.h"
#include "display.h"

/* Forward declarations */

/* Displays a list of items headed by *list, each indented with the string
   *indent */
static int fdisplay_list(FILE *f, const thing_data *list,
                         const display_fmt fmt, const char* indent,
                         const char* title);

/* Main workhorse of all display functions.  Prints out information on *thing
   in the given format and indented with the string *indent. */
static int do_display(FILE *f, const thing_data *thing,
                      const display_fmt fmt, const char* indent);

/* Implementation */

/* The first few functions are just wrappers that pass the real work off to
   other functions */
   
int display_thing(const thing_data *thing, const display_fmt fmt) {
  return do_display(stdout, thing, fmt, "");
}

int fdisplay_thing(FILE *f, const thing_data *thing, const display_fmt fmt) {
  return do_display(f, thing, fmt, "");
}

int display_everything(const display_fmt fmt) {
  fdisplay_everything(stdout, fmt);
}

/* Now a couple wrappers for dealing with groupd of Things */

int fdisplay_everything(FILE *f, const display_fmt fmt) {
  int result = 0;
  thing_type type;
  char title[32];

  for (type = tt_min ; type <= tt_max ; type++) {
    sprintf(title, "\n# %s\n", thing_name[type]);
    result += fdisplay_list(f, find_thing(type, ""), fmt, "", title);
  }

  return result;
}

static int fdisplay_list(FILE *f, const thing_data *list,
                         const display_fmt fmt, const char* indent,
                         const char* title) {
  const thing_data *curr = list;
  int result = 0;

  if (curr) fprintf(f, title);

  while (curr) {
    result += do_display(f, curr, fmt, indent);
    curr = next_thing(curr);
  }

  return result;
}

/* And, finally, do_display - the real meat of the module */

static int do_display(FILE *f, const thing_data *thing,
                      const display_fmt fmt, const char* indent) {
  int result = 1;
  char new_indent[64];
  char sub_indent[64];
  char tmpstr[32], tmpstr2[32];

  if (!thing) return 0;

  /* We don't have any standard header data here, so go straight into
     checking *thing's type */
  switch (thing->type) {
    case tt_unknown: {
      fprintf(f, "%s%s (type unknown)\n", indent, thing->data.generic.name);
      
      break;
    }

    case tt_resource_template: {
      fprintf(f, "%sResource %s", indent, thing->data.resource_template.name);

      switch (fmt) {
        case df_short: { fprintf(f, "\n"); break; }

        case df_long: {
          sprintf(new_indent, "%s  ", indent);
          sprintf(sub_indent, "%s  ", new_indent);

          fprintf(f, ":");
          if (thing->data.resource_template.can_overspend)
            fprintf(f, "  Can overspend by %d",
                    thing->data.resource_template.can_overspend);
          fprintf(f, "\n");

          if (thing->data.resource_template.influence) {
            sprintf(tmpstr, "%sInfluences:\n", new_indent);
            fdisplay_list(f, thing->data.resource_template.influence, fmt,
                          sub_indent, tmpstr);
          }

          if (thing->data.resource_template.decay_rate) {
            switch (thing->data.resource_template.decay_when) {
              case dt_early:  { strcpy(tmpstr, "early");  break; }
              default:        { strcpy(tmpstr, "late");   break; }  
            }
            sprintf(tmpstr2, "%s%d decays %s to:\n", new_indent,
                    thing->data.resource_template.decay_rate, tmpstr);
            fdisplay_list(f, thing->data.resource_template.decay_to, fmt,
                          sub_indent, tmpstr2);
          }

          break;
        }
      }

      break;
    }
    
    case tt_resource: {
      if (thing->data.resource.qty == 0) return 0;

      fprintf(f, "%s%d %s", indent, thing->data.resource.qty,
              thing->data.resource.name);
      if (strcmp(thing->data.resource.name, thing->data.resource.type->name))
        fprintf(f, "(%s)", thing->data.resource.type->name);
      fprintf(f, "\n");

      break;
    }
    
    case tt_race: {
      fprintf(f, "%sRace %s", indent, thing->data.race.name);

      switch (fmt) {
        case df_short: { fprintf(f, "\n"); break; }

        case df_long: {
          sprintf(new_indent, "%s  ", indent);
          sprintf(sub_indent, "%s  ", new_indent);

          fprintf(f, ":  Growth %d, Wanderlust %d\n",
                  thing->data.race.growth_rate, thing->data.race.wanderlust);
                  
          fprintf(f, "%sConsumes %d Food, %d Mana", new_indent,
                  thing->data.race.food_consumption,
                  thing->data.race.mana_consumption);
          if (thing->data.race.resource_consumption) 
            fdisplay_list(f, thing->data.race.resource_consumption, fmt,
                          sub_indent, ", other:\n");
          else
            fprintf(f, "\n");

          fprintf(f, "%sProduces %d Might, %d Labor, %d Knowledge", new_indent,
                  thing->data.race.might_production,
                  thing->data.race.labor_production,
                  thing->data.race.knowledge_production);
          if (thing->data.race.resource_production)
            fdisplay_list(f, thing->data.race.resource_production, fmt,
                          sub_indent, ", other:\n");
          else
            fprintf(f, "\n");

          break;
        }
      }
      
      break;
    }
    
    case tt_population: {
      fprintf(f, "%s%d %s", indent, thing->data.population.size,
              thing->data.population.name);
      if (strcmp(thing->data.population.name,
          thing->data.population.race->name))
        fprintf(f, " (%s)", thing->data.population.race->name);
      fprintf(f, "\n");

      if (fmt == df_long) {
        sprintf(new_indent, "%s  ", indent);
        fprintf(f, "%sSkill %d, Happiness %d\n", new_indent,
                thing->data.population.skill,
                thing->data.population.happiness);
      }

      break;
    }
    
    case tt_factory_template: {
      fprintf(f, "%sFactory %s", indent, thing->data.factory_template.name);

      switch (fmt) {
        case df_short: { fprintf(f, "\n"); break; }

        case df_long: {
          sprintf(new_indent, "%s  ", indent);
          sprintf(sub_indent, "%s  ", new_indent);

          fprintf(f, ":  Capacity %d, Depletion %d\n",
                  thing->data.factory_template.capacity,
                  thing->data.factory_template.depletion);

          sprintf(tmpstr, "%sInput:\n", new_indent);
          fdisplay_list(f, thing->data.factory_template.input, fmt,
                        sub_indent, tmpstr);

          sprintf(tmpstr, "%sOutput:\n", new_indent);
          fdisplay_list(f, thing->data.factory_template.output, fmt,
                        sub_indent, tmpstr);

          sprintf(tmpstr, "%sCost:\n", new_indent);
          fdisplay_list(f, thing->data.factory_template.cost, fmt,
                        sub_indent, tmpstr);

          sprintf(tmpstr, "%sRequires:\n", new_indent);
          fdisplay_list(f, thing->data.factory_template.required, fmt,
                        sub_indent, tmpstr);

          sprintf(tmpstr, "%sRepair cost per Capacity:\n", new_indent);
          fdisplay_list(f, thing->data.factory_template.repair_cost, fmt,
                        sub_indent, tmpstr);

          break;
        }
      }
      
      break;
    }
    
    case tt_factory: {
      fprintf(f, "%s%d %s", indent, thing->data.factory.qty,
              thing->data.factory.name);
      if (strcmp(thing->data.factory.name, thing->data.factory.type->name))
        fprintf(f, " (%s)", thing->data.factory.type->name);
      fprintf(f, ", Capacity %d\n", thing->data.factory.capacity);

      break;
    }
    
    case tt_site_template: {
      fprintf(f, "%sSite %s\n", indent, thing->data.site_template.name);

      if (fmt == df_long) { 
        sprintf(new_indent, "%s  ", indent);
        sprintf(sub_indent, "%s  ", new_indent);

        sprintf(tmpstr, "%sAffects with Magnitude %d:\n", new_indent,
                thing->data.site_template.magnitude);
        fdisplay_list(f, thing->data.site_template.affects, fmt,
                      sub_indent, tmpstr);
      }

      break;
    }
    
    case tt_site: {
      fprintf(f, "%s%s", indent, thing->data.site.name);
      if (strcmp(thing->data.site.name, thing->data.site.type->name))
        fprintf(f, " (%s)", thing->data.site.type->name);
      fprintf(f, "\n");

      break;
    }
    
    case tt_allocation: {
      switch (thing->data.allocation.type) {
        case al_quantity  : { strcpy(tmpstr, "by quantity");  break; }
        case al_percent   : { strcpy(tmpstr, "by percent");   break; }
        default           : { strcpy(tmpstr, "maximum");      break; }
      }

      sprintf(new_indent, "%s  ", indent);
      fprintf(f, "%sAllocation for %s %s", indent,
              thing->data.allocation.name, tmpstr);
      if ((fmt == df_long) && thing->data.allocation.resource) {
        fprintf(f, ":\n");
        fdisplay_list(f, thing->data.allocation.resource, fmt, new_indent, "");
      } else
        fprintf(f, "\n");

      break;
    }
    
    case tt_construction: {
      fprintf(f, "%sBuilding %s", indent, thing->data.construction.name);
      if (strcmp(thing->data.construction.name, 
          thing->data.construction.factory->name))
        fprintf(f, " (%s)", thing->data.construction.factory->name);
      fprintf(f, "\n");

      if (fmt == df_long) {
        sprintf(new_indent, "%s  ", indent);
        sprintf(sub_indent, "%s  ", new_indent);

        sprintf(tmpstr, "%sRemaining cost:\n", new_indent,
                thing->data.site_template.magnitude);
        fdisplay_list(f, thing->data.construction.remaining_cost, fmt,
                      sub_indent, tmpstr);
      }

      break;
    }
    
    case tt_territory: {
      sprintf(new_indent, "%s  ", indent);
      sprintf(sub_indent, "%s  ", new_indent);

      fprintf(f, "%sTerritory %s\n", indent, thing->data.territory.name);

      sprintf(tmpstr, "%sPopulace:\n", new_indent);
      fdisplay_list(f, thing->data.territory.populace, fmt, sub_indent, tmpstr);

      sprintf(tmpstr, "%sSites:\n", new_indent);
      fdisplay_list(f, thing->data.territory.sites, fmt, sub_indent, tmpstr);

      sprintf(tmpstr, "%sFactories:\n", new_indent);
      fdisplay_list(f, thing->data.territory.factories, fmt, sub_indent,
                    tmpstr);

      sprintf(tmpstr, "%sResources:\n", new_indent);
      fdisplay_list(f, thing->data.territory.resources, fmt, sub_indent,
                    tmpstr);

      sprintf(tmpstr, "%sAllocations:\n", new_indent);
      fdisplay_list(f, thing->data.territory.allocated, fmt, sub_indent,
                    tmpstr);

      sprintf(tmpstr, "%sConstruction:\n", new_indent);
      fdisplay_list(f, thing->data.territory.building, fmt, sub_indent, tmpstr);

      break;
    }

    default: {
      /* Should never be here */
      fprintf(f, "%sEncoutered thing with invalid type %d!", indent,
              thing->type);
    }
  }
}
