/* Core library for the Barons strategic engine */

#include <stdio.h>
#include "dslib.h"
#include "core.h"
#include "produce.h"

/* Internal utility functions */

static resource_data *get_res(territory_data *territory,
                              const char* resource_name) {
  resource_data *res;
  thing_data *thing;

  thing = find_in_list(territory->resources, resource_name);
  if (thing) return &(thing->data.resource);

  /* Create resource record in list */
  thing = (thing_data*)calloc(1, sizeof(thing_data));
  if (!thing) return 0;

  thing->type = tt_resource;
  res = &(thing->data.resource);
  strcpy(res->name, resource_name);
  res->type = &find_any_thing(resource_name)->data.resource_template;
  res->qty = 0;

  add_to_list(&(territory->resources), thing);

  return res;
}

static allocation_data *get_alloc(territory_data *territory,
                                  const char* consumer) {
  allocation_data *alloc;
  thing_data *thing;

  thing = find_in_list(territory->allocated, consumer);
  if (thing) return &(thing->data.allocation);

  /* Create default allocation record */
  thing = (thing_data*)calloc(1, sizeof(thing_data));
  if (!thing) return 0;

  thing->type = tt_allocation;
  alloc = &(thing->data.allocation);
  strcpy(alloc->name, consumer);
  alloc->type = al_max;

  add_to_list(&(territory->allocated), thing);

  return alloc;
}

static long calc_avail(const long desired_transactions,
                       const long transaction_size, territory_data *territory,
                       const char* resource_name, const char* consumer) {
  resource_data *res;
  allocation_data *alloc;
  thing_data *thing;
  long avail;
  long requested;
  long max_spend;
  long spend;

  res = get_res(territory, resource_name);
  alloc = get_alloc(territory, consumer);
  avail = res->qty;
  requested = desired_transactions * transaction_size;
  max_spend = avail 
            + round_rand(1.0 * res->type->can_overspend * res->old_qty
                         / res_scale);

  switch (alloc->type) {
    case al_quantity : {
      thing = find_in_list(alloc->resource, resource_name);
      
      if (thing) spend = thing->data.resource.qty;
      else spend = avail;

      break;
    }

    case al_percent : {
      thing = find_in_list(alloc->resource, resource_name);
      
      if (thing) spend = round_rand(1.0 * thing->data.resource.qty / 100);
      else spend = avail;

      break;
    }

    case al_max : {
      spend = max_spend;
    }
  }

  if (spend > max_spend) spend = max_spend;

  /* Supplies are sufficient to complete all requested transactions */
  if (spend >= requested) return desired_transactions;

  printf("Insufficient supply of %s for %s!\n", resource_name, consumer);
  printf("     Will spend %ld of %ld, Want %ld\n", spend, avail, requested);

  /* Can't fulfill request, so return the most we can manage */
  return (spend > 0) ? (spend / transaction_size) : 0;
}

/* ---------------------------------------------------------------------- */

/* Turn processing */

/* Overseer which just calls all the others */
void process_turn(territory_data* territory) {
  consume(territory);
  produce(territory);
  decay(territory);
}

static void do_decay(territory_data* territory, decay_type d_type);

/* Feed the people!  (And get Might, Labor, and Knowledge out of them...)  */
void consume(territory_data* territory) {
  thing_data *pop_rec;
  population_data *pop;
  race_data *race;

  thing_data *res;
  resource_data *supply;
  long can_feed;
  double morale;
  double productivity;

  long tmpint;
  double scale;

  /* Iterate through the populations in this territory */
  pop_rec = find_in_list(territory->populace, "");
  while (pop_rec) {
    pop = &(pop_rec->data.population);
    race = pop->race;
    if (!(can_feed = pop->size)) {    /* Skip if pop size is 0 */
      pop_rec = next_thing(pop_rec);
      continue;
    }
    
    /* Determine how many of this population can be fed */
    if (race->food_consumption)
      can_feed = calc_avail(can_feed, race->food_consumption, territory,
                            "Food", pop->name);

    if (race->mana_consumption)
      can_feed = calc_avail(can_feed, race->mana_consumption, territory,
                            "Mana", pop->name);

    if (race->resource_consumption) {
      res = find_in_list(race->resource_consumption, "");
      while (res) {
        can_feed = calc_avail(can_feed, res->data.resource.qty, territory,
                              res->data.resource.name, pop->name);
        res = next_thing(res);
      }
    }

    /* Subtract supplies consumed */
    if (race->food_consumption)
      get_res(territory, "Food")->qty -= can_feed * race->food_consumption;

    if (race->mana_consumption)
      get_res(territory, "Mana")->qty -= can_feed * race->mana_consumption;

    if (race->resource_consumption) {
      res = find_in_list(race->resource_consumption, "");
      while (res) {
        get_res(territory, res->data.resource.name)->qty -=
                  can_feed * res->data.resource.qty;
        res = next_thing(res);
      }
    }

    /* Handle early decay */
    do_decay(territory, dt_early);

    printf("Can feed %ld of %ld %s population\n", can_feed, pop->size, 
           pop->name);

    /* Adjust pop happiness, skill, and size */
    supply = get_res(territory, "Happiness");
    tmpint = supply->qty; supply->qty = 0;
    pop->happiness = round_rand((pop->happiness * 0.75) 
                              + (tmpint / pop->size));

    supply = get_res(territory, "Skill");
    tmpint = supply->qty; supply->qty = 0;
    pop->skill = round_rand((pop->skill * 0.9) + (tmpint / pop->size));

    if (can_feed < pop->size) {
      /* The people are starving! */
      tmpint = pop->size - can_feed;
      pop->happiness -= round_rand(500.0 * tmpint / can_feed);
      /* Multiplier to skill shifts from survival rate towards 1 because we
         assume that the more skilled members of the population will be more
         likely to survive */
      pop->skill = round_rand(pop->skill * 0.25 
                              * (3 + (can_feed / pop->size)));
      pop->size = can_feed;
    }

    morale = 1.0 + (pop->happiness * 1.0 / res_scale);
    productivity = 1.0 + (((3.0 * pop->skill) + (2.0 * pop->happiness))
                            / (5.0 * res_scale));
    printf("Morale: %f, Productivity: %f\n", morale, productivity);
    pop->size = round_rand(pop->size * (1.0 
                           + (race->growth_rate * morale / res_scale)));
    printf("%s population is now %ld\n", pop->name, pop->size);
    can_feed *= productivity;  /* Avoid doing this multiply repeatedly */

    /* Add resources produced */
    get_res(territory, "Might")->qty += can_feed * race->might_production;
    get_res(territory, "Labor")->qty += can_feed * race->labor_production;
    get_res(territory, "Knowledge")->qty += 
                                   can_feed * race->knowledge_production;

    if (race->resource_production) {
      res = find_in_list(race->resource_production, "");
      while (res) {
        get_res(territory, res->data.resource.name)->qty +=
                  can_feed * res->data.resource.qty;
        res = next_thing(res);
      }
    }

    pop_rec = next_thing(pop_rec);
  }
}

/* Make them work!  Turn the Labor into something useful! */
void produce(territory_data* territory) {
  thing_data *curr_rec;
  long res_qty;
  short can_produce;
  thing_data *res_rec;
  int i;

  /* Capture initial resource quantities */
  curr_rec = find_in_list(territory->resources, "");
  while (curr_rec) {
    curr_rec->data.resource.old_qty = curr_rec->data.resource.qty;
    curr_rec = next_thing(curr_rec);
  }

  /* Process influence for all resources in territory */
  curr_rec = find_in_list(territory->resources, "");
  while (curr_rec) {
    if (curr_rec->data.resource.type->influence) {
      res_qty = curr_rec->data.resource.old_qty;
      res_rec = find_in_list(curr_rec->data.resource.type->influence, "");
      while (res_rec) {
        get_res(territory, res_rec->data.generic.name)->qty
                  += round_rand(1.0 * res_rec->data.resource.qty * res_qty
                     / res_scale);
        
        res_rec = next_thing(res_rec);
      }
    }
    
    curr_rec = next_thing(curr_rec);
  }

  /* Process effects of all sites in territory */
  curr_rec = find_in_list(territory->sites, "");
  while (curr_rec) {
    if (curr_rec->data.site.type->affects) {
      res_rec = find_in_list(curr_rec->data.site.type->affects, "");
      while (res_rec) {
        get_res(territory, res_rec->data.generic.name)->qty
                  *= round_rand(1.0 * res_rec->data.resource.qty / res_scale);

        res_rec = next_thing(res_rec);
      }
    }
  
    curr_rec = next_thing(curr_rec);
  }

  /* Feed resources through factories */
  curr_rec = find_in_list(territory->factories, "");
  while (curr_rec) {
    /* Figure out how many transactions we can put through this factory */
    can_produce = curr_rec->data.factory.capacity;

    res_rec = find_in_list(curr_rec->data.factory.type->input, "");
    while (res_rec) {
      can_produce = calc_avail(can_produce, res_rec->data.resource.qty,
                               territory, res_rec->data.resource.name,
                               curr_rec->data.factory.name);
      res_rec = next_thing(res_rec);
    }

    /* Subtract supplies consumed */
    res_rec = find_in_list(curr_rec->data.factory.type->input, "");
    while (res_rec) {
      get_res(territory, res_rec->data.resource.name)->qty -=
                  can_produce * res_rec->data.resource.qty;
      res_rec = next_thing(res_rec);
    }

    /* Stockpile the factory's output */
    res_rec = find_in_list(curr_rec->data.factory.type->output, "");
    while (res_rec) {
      get_res(territory, res_rec->data.resource.name)->qty +=
                  can_produce * res_rec->data.resource.qty;
      res_rec = next_thing(res_rec);
    }

    /* Check for factory depletion */
    /*** CODE ME LATER ***/
  
    curr_rec = next_thing(curr_rec);
  }
}

/* Watch the surplus disappear!  (Hopefully turning into something useful in
   the process...) */
void do_decay(territory_data* territory, decay_type d_type) {
  thing_data *res_rec;
  resource_data *res;
  thing_data *target;
  long loss;

  res_rec = find_in_list(territory->resources, "");
  while (res_rec) {
    res = &(res_rec->data.resource);
    if ((res->type->decay_rate) && (res->type->decay_when == d_type)) {
      loss = round_rand(1.0 * res->qty * res->type->decay_rate / 100);
      res->qty -= loss;

      if (res->type->decay_to) {
        target = find_in_list(res->type->decay_to, "");
        while(target) {
          get_res(territory, target->data.generic.name)->qty
                    += round_rand(1.0 * target->data.resource.qty * loss
                       / res_scale);
          target = next_thing(target);
        }
      }
    }
    
    res_rec = next_thing(res_rec);
  }
}

void decay(territory_data* territory) {
  return do_decay(territory, dt_late);
}

